;;; 
;;;	Universal Monitor for Toshiba TLCS-90
;;;

	CPU	90c141

TARGET:	equ	"90C802"


	INCLUDE	"config.inc"

	INCLUDE	"../common.inc"

	
;;; 
;;; ROM area
;;;

	ORG	0000H
	DI
	JP	CSTART

	;;
	;; Entry point
	;;

	ORG	ENTRY+0		; Cold start
E_CSTART:
	JP	CSTART
	
	ORG	ENTRY+8		; Warm start
E_WSTART:
	JP	WSTART

	ORG	ENTRY+16	; Console output
E_CONOUT:
	JP	CONOUT

	ORG	ENTRY+24	; (Console) String output
E_STROUT:
	JP	STROUT

	ORG	ENTRY+32	; Console input
E_CONIN:
	JP	CONIN

	ORG	ENTRY+40	; Console status
E_CONST:
	JP	CONST

	;;
	;;
	;; 

	ORG	0100H
CSTART:
	;; Disable WDT
	XOR	A,A
	LD	(0FFD2H),A
	LD	A,0B1H
	LD	(0FFD3H),A

	LD	SP,STACK

	CALL    INIT

	LD	HL,8000H
	LD	(DSADDR),HL
	LD	(SADDR),HL
	LD	(GADDR),HL
	LD	A,'I'
	LD	(HEXMOD),A
	
	;; Opening message
	LD	HL,OPNMSG
	CALL	STROUT

WSTART:
	LD	HL,PROMPT
	CALL	STROUT
	CALL	GETLIN
	LD	HL,INBUF
	CALL	SKIPSP
	CALL	UPPER
	OR	A,A
	JR	Z,WSTART

	CP	A,'D'
	JR	Z,DUMP
	CP	A,'G'
	JP	Z,GO
	CP	A,'S'
	JP	Z,SETM

	CP	A,'L'
	JP	Z,LOADH
	CP	A,'P'
	JP	Z,SAVEH

	CP	A,'I'
	JP	Z,PIN
	CP	A,'O'
	JP	Z,POUT

ERR:
	LD	HL,ERRMSG
	CALL	STROUT
	JR	WSTART

;;; 
;;; Dump memory
;;; 
DUMP:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX		; 1st arg.
	LD	A,C
	OR	A,A
	JR	NZ,DP0
	;; No arg.
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JR	NZ,ERR
	LD	HL,(DSADDR)
	LD	BC,128
	ADD	HL,BC
	LD	(DEADDR),HL
	JR	DPM

	;; 1st arg. found
DP0:
	LD	(DSADDR),DE
	CALL	SKIPSP
	LD	A,(HL)
	CP	A,','
	JR	Z,DP1
	OR	A,A
	JR	NZ,ERR
	;; No 2nd arg.
	LD	HL,128
	ADD	HL,DE
	LD	(DEADDR),HL
	JR	DPM
DP1:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	LD	A,C
	OR	A,A
	JR	Z,ERR
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR
	INC	DE
	LD	(DEADDR),DE
DPM:
	;; DUMP main
	LD	HL,(DSADDR)
	LD	A,0F0H
	AND	A,L
	LD	L,A
	XOR	A,A
	LD	(DSTATE),A
DPM0:
	PUSH	HL
	CALL	DPL
	POP	HL
	LD	BC,16
	ADD	HL,BC
	CALL	CONST
	OR	A,A
	JR	NZ,DPM1
	LD	A,(DSTATE)
	CP	A,2
	JR	C,DPM0
	LD	HL,(DEADDR)
	LD	(DSADDR),HL
	JP	WSTART
DPM1:
	LD	(DSADDR),HL
	CALL	CONIN
	JP	WSTART

DPL:
	;; DUMP line
	CALL	HEXOUT4
	PUSH	HL
	LD	HL,DSEP0
	CALL	STROUT
	POP	HL
	LD	IX,INBUF
	LD	B,16
DPL0:
	CALL	DPB
	DJNZ	DPL0

	LD	HL,DSEP1
	CALL	STROUT

	LD	HL,INBUF
	LD	B,16
DPL1:
	LD	A,(HL)
	INC	HL
	CP	A,' '
	JR	C,DPL2
	CP	A,7FH
	JR	NC,DPL2
	CALL	CONOUT
	JR	DPL3
DPL2:
	LD	A,'.'
	CALL	CONOUT
DPL3:
	DJNZ	DPL1
	JP	CRLF

DPB:	; Dump byte
	LD	A,' '
	CALL	CONOUT
	LD	A,(DSTATE)
	OR	A,A
	JR	NZ,DPB2
	; Dump state 0
	LD	A,(DSADDR)	; Low byte
	CP	A,L
	JR	NZ,DPB0
	LD	A,(DSADDR+1)	; High byte
	CP	A,H
	JR	Z,DPB1
DPB0:	; Still 0 or 2
	LD	A,' '
	CALL	CONOUT
	LD	A,' '
	CALL	CONOUT
	LD	(IX),A
	INC	HL
	INC	IX
	RET
DPB1:	; Found start address
	LD	A,1
	LD	(DSTATE),A
DPB2:
	LD	A,(DSTATE)
	CP	A,1
	JR	NZ,DPB0
	; Dump state 1
	LD	A,(HL)
	LD	(IX),A
	CALL	HEXOUT2
	INC	HL
	INC	IX
	LD	A,(DEADDR)	; Low byte
	CP	A,L
	RET	NZ
	LD	A,(DEADDR+1)	; High byte
	CP	A,H
	RET	NZ
	; Found end address
	LD	A,2
	LD	(DSTATE),A
	RET

;;;
;;; GO address
;;; 
GO:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR
	LD	A,C
	OR	A,A
	JR	Z,G0
	LD	(GADDR),DE
G0:
	LD	HL,(GADDR)
	JP	(HL)

;;;
;;; SET memory
;;; 
SETM:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR
	LD	A,C
	OR	A,A
	JR	NZ,SM0
	LD	DE,(SADDR)
SM0:
	EX	DE,HL
SM1:
	CALL	HEXOUT4
	PUSH	HL
	LD	HL,DSEP1
	CALL	STROUT
	POP	HL
	LD	A,(HL)
	PUSH	HL
	CALL	HEXOUT2
	LD	A,' '
	CALL	CONOUT
	CALL	GETLIN
	LD	HL,INBUF
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JR	NZ,SM2
	;; Empty  (Increment address)
	POP	HL
	INC	HL
	LD	(SADDR),HL
	JR	SM1
SM2:
	CP	A,'-'
	JR	NZ,SM3
	;; '-'  (Decrement address)
	POP	HL
	DEC	HL
	LD	(SADDR),HL
	JR	SM1
SM3:
	CP	A,'.'
	JR	NZ,SM4
	POP	HL
	LD	(SADDR),HL
	JP	WSTART
SM4:
	CALL	RDHEX
	LD	A,C
	OR	A,A
	POP	HL
	JP	Z,ERR
	LD	(HL),E
	INC	HL
	LD	(SADDR),HL
	JR	SM1

;;;
;;; LOAD HEX file
;;;
LOADH:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR

	LD	A,C
	OR	A,A
	JR	NZ,LH0

	LD	DE,0		;Offset
LH0:
	CALL	CONIN
	CALL	UPPER
	CP	A,'S'
	JR	Z,LHS0
LH1:
	CP	A,':'
	JR	Z,LHI0
LH2:
	;; Skip to EOL
	CP	A,CR
	JR	Z,LH0
	CP	A,LF
	JR	Z,LH0
LH3:
	CALL	CONIN
	JR	LH2

LHI0:
	CALL	HEXIN
	LD	C,A		; Checksum
	LD	B,A		; Length

	CALL	HEXIN
	LD	H,A		; Address H
	ADD	A,C
	LD	C,A

	CALL	HEXIN
	LD	L,A		; Address L
	ADD	A,C
	LD	C,A

	;; Add offset
	ADD	HL,DE

	CALL	HEXIN
	LD	(RECTYP),A
	ADD	A,C
	LD	C,A		; Checksum

	LD	A,B
	OR	A,A
	JR	Z,LHI3
LHI1:
	CALL	HEXIN
	PUSH	AF
	ADD	A,C
	LD	C,A		; Checksum

	LD	A,(RECTYP)
	OR	A,A
	JR	NZ,LHI20

	POP	AF
	LD	(HL),A
	INC	HL
	JR	LHI2
LHI20:
	POP	AF
LHI2:
	DJNZ	LHI1
LHI3:
	CALL	HEXIN
	ADD	A,C
	JR	NZ,LHIE		; Checksum error
	LD	A,(RECTYP)
	OR	A,A
	JP	Z,LH3
	JP	WSTART
LHIE:
	LD	HL,IHEMSG
	CALL	STROUT
	JP	WSTART
	
LHS0:
	CALL	CONIN
	LD	(RECTYP),A

	CALL	HEXIN
	LD	B,A		; Length+3
	LD	C,A		; Checksum

	CALL	HEXIN
	LD	H,A
	ADD	A,C
	LD	C,A
	
	CALL	HEXIN
	LD	L,A
	ADD	A,C
	LD	C,A

	ADD	HL,DE

	DEC	B
	DEC	B
	DEC	B
	JR	Z,LHS3
LHS1:
	CALL	HEXIN
	PUSH	AF
	ADD	A,C
	LD	C,A		; Checksum

	LD	A,(RECTYP)
	CP	A,'1'
	JR	NZ,LHS2

	POP	AF
	LD	(HL),A
	INC	HL
	JR	LHS20
LHS2:
	POP	AF
LHS20:
	DJNZ	LHS1
LHS3:
	CALL	HEXIN
	ADD	A,C
	CP	A,0FFH
	JR	NZ,LHSE

	LD	A,(RECTYP)
	CP	A,'7'
	JR	Z,LHSR
	CP	A,'8'
	JR	Z,LHSR
	CP	A,'9'
	JR	Z,LHSR
	JP	LH3
LHSE:
	LD	HL,SHEMSG
	CALL	STROUT
LHSR:
	JP	WSTART
	
;;;
;;; SAVE HEX file
;;;
SAVEH:
	INC	HL
	LD	A,(HL)
	CALL	UPPER
	CP	A,'I'
	JR	Z,SH0
	CP	A,'S'
	JR	NZ,SH1
SH0:
	INC	HL
	LD	(HEXMOD),A
SH1:
	CALL	SKIPSP
	CALL	RDHEX
	LD	A,C
	OR	A,A
	JR	Z,SHE
	PUSH	DE
	POP	IX		; IX = Start address
	CALL	SKIPSP
	LD	A,(HL)
	CP	A,','
	JR	NZ,SHE
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX		; DE = End address
	LD	A,C
	OR	A,A
	JR	Z,SHE
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JR	Z,SH2
SHE:
	JP	ERR

SH2:
	PUSH	IX
	POP	HL
	EX	DE,HL
	INC	HL
	OR	A,A
	SBC	HL,DE		; HL = Length
SH3:
	CALL	SHL
	LD	A,H
	CP	A,L
	JR	NZ,SH3

	LD	A,(HEXMOD)
	CP	A,'I'
	JR	NZ,SH4
	;; End record for Intel HEX
	LD	HL,IHEXER
	CALL	STROUT
	JP	WSTART
SH4:
	;; End record for Motorola S record
	LD	HL,SRECER
	CALL	STROUT
	JP	WSTART

SHL:
	LD	C,16
	LD	A,H
	OR	A,A
	JR	NZ,SHL0
	LD	A,L
	CP	A,C
	JR	NC,SHL0
	LD	C,A
SHL0:
	LD	B,0
	OR	A,A
	SBC	HL,BC
	LD	B,C

	LD	A,(HEXMOD)
	CP	A,'I'
	JR	NZ,SHLS

	;; Intel HEX
	LD	A,':'
	CALL	CONOUT

	LD	A,B
	CALL	HEXOUT2		; Length
	LD	C,B		; Checksum

	LD	A,D
	CALL	HEXOUT2
	LD	A,D
	ADD	A,C
	LD	C,A
	
	LD	A,E
	CALL	HEXOUT2
	LD	A,E
	ADD	A,C
	LD	C,A
	
	XOR	A,A
	CALL	HEXOUT2
SHLI0:
	LD	A,(DE)
	PUSH	AF
	CALL	HEXOUT2
	POP	AF
	ADD	A,C
	LD	C,A

	INC	DE
	DJNZ	SHLI0

	LD	A,C
	NEG	A
	CALL	HEXOUT2
	JP	CRLF

SHLS:
	;; Motorola S record
	LD	A,'S'
	CALL	CONOUT
	LD	A,'1'
	CALL	CONOUT

	LD	A,B
	ADD	A,2+1		; DataLength + 2(Addr) + 1(Sum)
	LD	C,A
	CALL	HEXOUT2

	LD	A,D
	CALL	HEXOUT2
	LD	A,D
	ADD	A,C
	LD	C,A
	
	LD	A,E
	CALL	HEXOUT2
	LD	A,E
	ADD	A,C
	LD	C,A
SHLS0:
	LD	A,(DE)
	PUSH	AF
	CALL	HEXOUT2		; Data
	POP	AF
	ADD	A,C
	LD	C,A

	INC	DE
	DJNZ	SHLS0

	LD	A,C
	CPL	A
	CALL	HEXOUT2
	JP	CRLF

;;;
;;; Port in
;;; 
PIN:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	LD	A,C
	OR	A,A
	JP	Z,ERR		; Port no. missing
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR

	;; Byte
	LD	A,(DE)
	CALL	HEXOUT2
	CALL	CRLF
	JP	WSTART

;;;
;;; Port out
;;;
POUT:
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	LD	A,C
	OR	A,A
	JP	Z,ERR		; Port no. missing
	PUSH	DE
	POP	IX
	CALL	SKIPSP
	LD	A,(HL)
	CP	A,','
	JP	NZ,ERR
	INC	HL
	CALL	SKIPSP
	CALL	RDHEX
	LD	A,C
	OR	A,A
	JP	Z,ERR		; Data missing
	CALL	SKIPSP
	LD	A,(HL)
	OR	A,A
	JP	NZ,ERR

	PUSH	IX
	POP	BC

	;; Byte
	LD	(BC),E
	JP	WSTART

;;;
;;; Other support routines
;;;

STROUT:
	LD	A,(HL)
	AND	A,A
	RET	Z
	CALL	CONOUT
	INC	HL
	JR	STROUT

HEXOUT4:
	LD	A,H
	CALL	HEXOUT2
	LD	A,L
HEXOUT2:
	PUSH	AF
	RRA
	RRA
	RRA
	RRA
	CALL	HEXOUT1
	POP	AF
HEXOUT1:
	AND	A,0FH
	ADD	A,'0'
	CP	A,'9'+1
	JP	C,CONOUT
	ADD	A,'A'-'9'-1
	JP	CONOUT

HEXIN:
	XOR	A,A
	CALL	HI0
	RLCA
	RLCA
	RLCA
	RLCA
HI0:
	PUSH	BC
	LD	C,A
	CALL	CONIN
	CALL	UPPER
	CP	A,'0'
	JR	C,HIR
	CP	A,'9'+1
	JR	C,HI1
	CP	A,'A'
	JR	C,HIR
	CP	A,'F'+1
	JR	NC,HIR
	SUB	A,'A'-'9'-1
HI1:
	SUB	A,'0'
	OR	A,C
HIR:
	POP	BC
	RET
	
CRLF:
	LD	A,CR
	CALL	CONOUT
	LD	A,LF
	JP	CONOUT

GETLIN:
	LD	HL,INBUF
	LD	B,0
GL0:
	CALL	CONIN
	CP	A,CR
	JR	Z,GLE
	CP	A,LF
	JR	Z,GLE
	CP	A,BS
	JR	Z,GLB
	CP	A,DEL
	JR	Z,GLB
	CP	A,' '
	JR	C,GL0
	CP	A,80H
	JR	NC,GL0
	LD	C,A
	LD	A,B
	CP	A,BUFLEN-1
	JR	NC,GL0		; Too long
	INC	B
	LD	A,C
	LD	(HL),A
	CALL	CONOUT
	INC	HL
	JR	GL0
GLB:
	LD	A,B
	AND	A,A
	JR	Z,GL0
	DEC	B
	DEC	HL
	LD	A,08H
	CALL	CONOUT
	LD	A,' '
	CALL	CONOUT
	LD	A,08H
	CALL	CONOUT
	JR	GL0
GLE:
	CALL	CRLF
	LD	(HL),00H
	RET

SKIPSP:
	LD	A,(HL)
	CP	A,' '
	RET	NZ
	INC	HL
	JR	SKIPSP

UPPER:
	CP	A,'a'
	RET	C
	CP	A,'z'+1
	RET	NC
	ADD	A,'A'-'a'
	RET

RDHEX:
	LD	C,0
	LD	DE,0
RH0:
	LD	A,(HL)
	CALL	UPPER
	CP	A,'0'
	JR	C,RHE
	CP	A,'9'+1
	JR	C,RH1
	CP	A,'A'
	JR	C,RHE
	CP	A,'F'+1
	JR	NC,RHE
	SUB	A,'A'-'9'-1
RH1:
	SUB	A,'0'
	RLA
	RLA
	RLA
	RLA
	RLA
	RL	E
	RL	D
	RLA
	RL	E
	RL	D
	RLA
	RL	E
	RL	D
	RLA
	RL	E
	RL	D
	INC	HL
	INC	C
	JR	RH0
RHE:
	RET

OPNMSG:
	DB	CR,LF,"Universal Monitor TLCS-90",CR,LF,00H

PROMPT:
	DB	"] ",00H

IHEMSG:
	DB	"Error ihex",CR,LF,00H
SHEMSG:
	DB	"Error srec",CR,LF,00H
ERRMSG:
	DB	"Error",CR,LF,00H

DSEP0:
	DB	" :",00H
DSEP1:
	DB	" : ",00H
IHEXER:
        DB	":00000001FF",CR,LF,00H
SRECER:
        DB	"S9030000FC",CR,LF,00H

	IF USE_DEV_TLCS90
	INCLUDE	"dev/dev_tlcs90.asm"
	ENDIF

	IF USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF
	    
;;;
;;; RAM area
;;;

	;;
	;; Work Area
	;;
	
	ORG	WORK_B

INBUF:	DS	BUFLEN		; Line input buffer
DSADDR:	DS	2		; Dump start address
DEADDR:	DS	2		; Dump end address
DSTATE:	DS	1		; Dump state
GADDR:	DS	2		; Go address
SADDR:	DS	2		; Set address
HEXMOD:	DS	1		; HEX file mode
RECTYP:	DS	1		; Record type

	END
